from simpy import Environment
from typing import Callable
from simso.core.Processor import ProcInfo


class FaultFunction:
    """
    The FaultFunction class changes the actual behaviour of the faulty element.
    """

    def __init__(self, start_function: Callable, end_function: Callable):
        self.start_function = start_function
        self.end_function = end_function


class FaultInfo:
    """
    The FaultType class is responsible for causing the error.
    """

    def __init__(self, start_time: float, end_time: float, fault_function: FaultFunction):
        self.start_time = start_time
        self.end_time = end_time
        self.fault_function = fault_function


class Fault:
    """
    The Fault class is the simpy process for triggering a certain error.
    """

    def __init__(self, sim, fault_type: FaultInfo):
        self._fault_type = fault_type
        self._sim = sim
        self.process = self._sim.process(self.release())

    def release(self):
        yield self._sim.timeout(self._fault_type.start_time)

        self._fault_type.fault_function.start_function(self._sim)

        yield self._sim.timeout(self._fault_type.end_time -
                                self._fault_type.start_time)

        self._fault_type.fault_function.end_function(self._sim)


class ProcessorFailure(FaultFunction):
    """
    The ProcessorFailure class simulates a complete stoppage of a processor.
    """

    def __init__(self, proc_info: ProcInfo):
        super().__init__(self.stop_processor, self.start_processor)
        self._proc_info = proc_info
        self._processor = None

    def stop_processor(self, model):
        self._processor = next(
            p for p in model.processors if p.identifier == self._proc_info.identifier)
        self._processor.fail(self)

    def start_processor(self, model):
        self._processor.fail()


class VoltageChange(FaultFunction):

    def __init__(self, proc_info_list, scaling: float):
        super().__init__(self.start_voltage_change, self.end_voltage_change)
        self._proc_info_list = proc_info_list
        self._processor_list = []
        self._scaling = scaling

    def start_voltage_change(self, model):
        affected_ids = [p.identifier for p in self._proc_info_list]
        self._processor_list = [
            p for p in model.processors if p.identifier in affected_ids]
        for p in self._processor_list:
            p.set_speed(self._scaling * p.speed)

    def end_voltage_change(self, model):
        for p in self._processor_list:
            speed = next(
                pi for pi in self._proc_info_list if pi.identifier == p.identifier).speed
            p.set_speed(speed)


class VoltageDrop(VoltageChange):
    """
    The VoltageDrop class simulates the slowdown of one or more processors due to voltage drops.
    """

    def __init__(self, proc_info_list, scaling: float):
        assert scaling >= 0 and scaling <= 1.0, "Voltage drop reduces the speed of a processor, i.e. scaling has to be >= 0 and <= 1"
        super().__init__(proc_info_list, scaling)


# TODO: rethink this class
class PriorityInversion(FaultFunction):
    """
    The PriorityInversion class simulates changes in the priority of a job.
    """

    def __init__(self, proc_info, until_sched: bool = False):
        """
        lambd: Expects task and new value as parameter and returns old value
        """
        super().__init__(self.start_priority_inversion, self.end_priority_inversion)
        self._proc_info = proc_info
        self._until_sched = until_sched

    def start_priority_inversion(self, model):
        proc = next(p for p in model.processors if p.identifier ==
                    self._proc_info.identifier)
        sched = model.scheduler
        running_jobs = [
            p.running for p in model.processors if p.running is not None]
        job = next(t.job for t in sched.task_list if t.job.is_active()
                   and t.job not in running_jobs)

        if job is not None:
            pass

    def end_priority_inversion(self, model):
        pass


class TimingChange(FaultFunction):
    """
    The TimingError class simulates changes in the timing parameters, e.g. deadline, period etc.
    """

    def __init__(self, task_info, parameter: str, value: float):
        super().__init__(self.start_timing_change, self.end_timing_change)
        self._task_info = task_info
        self._parameter = parameter
        self._value = value
        self._old_value = getattr(self._task_info, parameter)

    def start_timing_change(self, model):
        setattr(self._task_info, self._parameter, self._value)

    def end_timing_change(self, model):
        setattr(self._task_info, self._parameter, self._old_value)


class TimingError(TimingChange):
    """
    The TimingError class simulates changes in the timing parameters, e.g. deadline, period etc.
    """

    def __init__(self, task_info, parameter: str, value: float):
        super().__init__(task_info, parameter, value)


class Redundancy(TimingChange):

    def __init__(self, task_info, value: float):
        super().__init__(task_info, "wcet", task_info.wcet + value)
